{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Computing enclosures for all global minimizers and for the global }
{    minimum value of a twice continuously differentiable multi-dimensional, }
{    scalar valued function, assuming that the global minimum is a stationa- }
{    ry point. If it is a boundary point of the search area with gradient of }
{    the function being different from zero, the method fails in its form    }
{    presented here.                                                         }
{    Note that some routines are implemented using the "modified call by     }
{    reference" of PASCAL-XSC (see Language Reference for details) to avoid  }
{    the very inefficient allocation of local memory for the copies of the   }
{    actual parameters.                                                      }
{ Method: Bisection method combined with midpoint, monotonicity, concavity   }
{    test and extended interval Newton step.                                 }
{ Global procedures and functions:                                           }
{    procedure AllGOp(...)       : computes enclosures for all zeros         }
{    function  AllGOpErrMsg(...) : delivers an error message text            }
{ Updates:                                                                   }
{    11.03.1994 : Use of new procedure 'Compress' to reduce the number of    }
{                 final boxes produced by the algorithm                      }
{    12.06.1995 : No 'p:=0' if NoSolution at the end of NewtonStep           }
{    23.11.1995 : Termination criterium changed to ' <= Epsilon'             }
{----------------------------------------------------------------------------}
module gop;

use
  i_ari,     { Interval arithmetic               }
  xi_ari,    { Extended interval arithmetic      }
  lst_ari,   { List arithmetic                   }
  i_util,    { Interval utilities                }
  b_util,    { Boolean utilities                 }
  hess_ari,  { Differentiation arithmetic        }
  mv_ari,    { Real matrix/vector arithmetic     }
  mvi_ari,   { Interval matrix/vector arithmetic }
  matinv,    { Inversion of real matrices        }
  mvi_util;  { Interval matrix/vector utilities  }

const
  ListSize = 1000;   { List size, i.e. number of vectors which can be stored }

{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError       = 0;  { No error occurred.                                   }
  lbOptiVecNot1 = 1;  { Lower bound of variable OptiVector is not equal to 1.}
  lbInfoVecNot1 = 2;  { Lower bound of variable InfoVector is not equal to 1.}
  VecsDiffer    = 3;  { Bounds of OptiVector and InfoVector do not match.    }
  VecTooSmall   = 4;  { OptiVector too small. Not all zeros can be stored.   }
  NoStatOpti    = 5;  { No stationary point is a global optimizer.           }
  ListTooSmall  = 6;  { Internal List is too small.                          }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function AllGOpErrMsg ( Err : integer ) : string;
var
  Msg : string;
begin
  case Err of
    NoError      : Msg := '';
    lbOptiVecNot1: Msg := 'Lower bound of OptiVector is not equal to 1';
    lbInfoVecNot1: Msg := 'Lower bound of InfoVector is not equal to 1';
    VecsDiffer   : Msg := 'Bounds of OptiVector and InfoVector do not match';
    VecTooSmall  : Msg := 'Not all optimizers found. OptiVector is too small';
    NoStatOpti   : Msg := 'No global optimizer (stationary point!) found';
    ListTooSmall : Msg := 'Internal list too small. ' +
                          'Increase "ListSize" and recompile "gop.p"!';
    else         : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  AllGOpErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Determine the component 'mc' with maximum interval diameter of 'iv[mc]'.   }
{----------------------------------------------------------------------------}
function MaxDiamComp (var iv: ivector) : integer;
var
  mc, k : integer;
  d     : rvector[lb(iv)..ub(iv)];
begin
  d:= diam(iv);  mc:= lb(iv);
  for k:=lb(iv)+1 to ub(iv) do
    if d[k] > d[mc] then mc:= k;
  MaxDiamComp:= mc;
end;

function MonotonicityTest (var gradY: ivector) : boolean; { returns true if  }
var                                                       { 'f' is monotone  }
  i, n   : integer;                                       {------------------}
  Delete : boolean;
begin
  Delete:= false; i:= 1;  n:= ub(gradY);
  while (i <= n) and (not Delete) do
  begin
    if (0 < gradY[i].inf) or (0 > gradY[i].sup) then   { 'f' is monotone }
      Delete:= true;                                   {-----------------}
    i:= i+1;
  end;
  MonotonicityTest:= Delete;
end;

function ConcavityTest (var HessY: imatrix) : boolean;  { returns true if    }
var                                                     { 'f' is not convex  }
  i, n   : integer;                                     {--------------------}
  Delete : boolean;
begin
  Delete:= false;  i:= 1;  n:= ub(HessY);
  while (i <= n) and (not Delete) do
  begin
    if (HessY[i,i].sup < 0) then  { 'f' is not convex }
      Delete:= true;              {-------------------}
    i:= i+1;
  end;
  ConcavityTest:= Delete;
end;

{----------------------------------------------------------------------------}
{ Purpose: Execution of one single interval Newton Gauss-Seidel step for the }
{    interval vector 'Y' and the gradient of 'f'.                            }
{ Parameters:                                                                }
{    In     : 'f'          : must be declared for the type 'HessType' to     }
{                            enable the internal use of the differentiation  }
{                            arithmetic 'hess_ari'.                          }
{             'Y'          : specifies the starting interval.                }
{             'HessY'      : Hessian matrix of 'f(Y)', already computed      }
{                            outside of 'NewtonStep'.                        }
{    Out    : 'V'          : stores the enclosures 'V[i]' for the splittings }
{                            generated by the Newton step.                   }
{             'p'          : number of non-empty interval vectors 'V[i]'.    }
{ Description:                                                               }
{    'NewtonStep' executes the extended interval Newton Gauss-Seidel step    }
{    for 'Y' with result interval vector(s) 'V[i]' which can be empty.       }
{    'p' gives the number of non-empty interval vectors stored as vectors    }
{    'V[1]',...,'V[p]'.                                                      }
{----------------------------------------------------------------------------}
procedure NewtonStep (function f (x: HTvector) : HessType[0..ub(x,1),
                                                          0..ub(x,2)];
                                             Y : ivector;
                           var           HessY : imatrix;
                           var               V : imatrix;
                           var               p : integer);
var
  c                        : rvector[1..ub(Y)];
  GradC, b, Y_minus_c      : ivector[1..ub(Y)];
  R                        : rmatrix[1..ub(Y),1..ub(Y)];
  A                        : imatrix[1..ub(Y),1..ub(Y)];
  i, i0, n, InvErr, j      : integer;
  NoSolution               : boolean;
  h, fC                    : interval;
  z                        : ivector[1..2];
begin
  c:= mid(Y);
  fgEvalH(f, intval(c), fC, GradC);         { Midpoint gradient evaluation   }
                                            {--------------------------------}
  MatInv(mid(HessY),R,InvErr);              { Invert the midpoint matrix     }
  if InvErr <> 0 then R:= id(R);            {--------------------------------}

  A:= R * HessY;  b:= R * GradC;        { Compute data for Gauss-Seidel step }
  Y_minus_c:= Y - c;                    {------------------------------------}

  p:= 0; i:= 0; i0:= 0; n:= ub(Y); { Initializations, A[i0,i0] contains zero }
                                   {-----------------------------------------}
  NoSolution:= false;
  while (i < n) and (not NoSolution) do     { Interval Gauss-Seidel step for }
  begin                                     { non-zero A[i,i] elements       }
    i:= succ (i);                           {--------------------------------}
    if not (0 in A[i,i]) then
      begin
        h:= ## (b[i] + (for j:=1 to i-1 sum (A[i,j] * Y_minus_c[j]) )
                     + (for j:=i+1 to n sum (A[i,j] * Y_minus_c[j]) ) );
        h:= c[i] - h / A[i,i];

        if (Y[i] >< h) then
          NoSolution:= true
        else
          begin
            Y[i]:= Y[i] ** h;  Y_minus_c[i]:= Y[i] - c[i];
          end;
      end
    else
      i0:= i;                                   { Largest i with 0 in A[i,i] }
  end;                                          {----------------------------}

  i:= 0;
  while (not NoSolution) and (i < i0) do    { Interval Gauss-Seidel step for }
  begin                                     { zero A[i,i] elements           }
    i:= succ (i);                           {--------------------------------}
    if (0 in A[i,i]) then
    begin
      h:= ## ( b[i] + ( for j:=1 to i-1 sum (A[i,j] * Y_minus_c[j]) )
                    + ( for j:=i+1 to n sum (A[i,j] * Y_minus_c[j]) ) );
      z:= Y[i] ** (c[i] - h div A[i,i]);        { Extended interval division }
                                                {----------------------------}
      if (z[1] = EmptyIntval) then              { z[1] = z[2] = EmptyIntval  }
        NoSolution:= true                       {----------------------------}
      else
        begin                                              { Compute new 'Y' }
          Y[i]:= z[1];  Y_minus_c[i]:= Y[i] - c[i];        {-----------------}
          if z[2] <> EmptyIntval then
          begin                                   { Store further bisections }
            p:= p+1; V[p]:= Y; V[p][i]:= z[2];    {--------------------------}
          end;
        end;
    end; { if 0 in A[i,i] ... }
  end; { while (not NoSolution) ... }

  if not NoSolution then
  begin
    p:= p+1;  V[p]:= Y;
  end;
end; { NewtonStep }


{----------------------------------------------------------------------------}
{ Purpose: Execution of the global optimization method including a bisection }
{    method, midpoint test, monotonicity test, concavity test, and extended  }
{    interval Newton steps.                                                  }
{ Parameters:                                                                }
{    In     : 'f'          : must be declared for type 'HTvector' to enable  }
{                            the internal use of the differentiation         }
{                            arithmetic 'hess_ari'.                          }
{             'Start:      : specifies the starting interval.                }
{             'Epsilon'    : specifies the desired relative accuracy.        }
{             'VecMat'     : matrix for storing the vectors of the list.     }
{    Out    : 'ResultList' : stores the candidates for enclosure of a global }
{                            minimizer.                                      }
{             'Minimum'    : stores the enclosure of the global minimum      }
{                            value.                                          }
{             'ListFull'   : signals that list (matrix VecMat) is full       }
{ Description:                                                               }
{   The procedure manages the list 'L' of pending subintervals that may      }
{   contain global minimizers. Subintervals are removed from the list and    }
{   placed in the accepted list 'ResultList' when they satisfy relative      }
{   error acceptance criteria. Subintervals are also removed from the list   }
{   by the midpoint, monotonicity, concavity tests, or by the interval       }
{   Newton steps. Subintervals are added to the pending list when an element }
{   from the list is bisected or when the extended interval Newton step      }
{   yields two candidate intervals.                                          }
{   'ResultList' returns the list of enclosures of the global minimizers,    }
{   'Minimum' returns the enclosure of the global minimum value.             }
{----------------------------------------------------------------------------}
procedure GlobalOptimize (function f (x: HTvector) : HessType[0..ub(x,1),
                                                              0..ub(x,2)];
                               var           Start : ivector;
                                           Epsilon : real;
                               var      ResultList : PairPtr;
                               var         Minimum : interval;
                               var          VecMat : imatrix;
                               var        ListFull : boolean);

var
  PairY               : ivector[0..ub(Start)];  { Pair ( Y, inf(f(Y) )       }
  Y                   : ivector[1..ub(Start)];
  U                   : imatrix[1..2,           { Subboxes of Y              }
                                1..ub(Start)];
  V                   : imatrix[1..ub(Start)+1, { Subboxes of U              }
                                1..ub(Start)];
  fY, fU, fV, fC, fCV : interval;               { Function evaluations of f  }
  gradU, gradV        : ivector[1..ub(Start)];  { Gradient evaluations of f  }
  HessU               : imatrix[1..ub(Start),   { Hessian evaluation of f    }
                                1..ub(Start)];
  fmax                : real;                   { Upper bound for minimum    }
  c, cV               : rvector[1..ub(Start)];  { Midpoints of Y and V       }
  WorkList            : PairPtr;                { List of pairs              }
  i, j, p, k          : integer;                { Control variables          }
  Bisect              : boolean;                { Flag for iteration         }

begin
  c:= mid(Start);
  fEvalH(f, intval(c), fC);                { Compute upper bound for minimum }
  fmax:= sup(fC);                          {---------------------------------}

  ResultList:= EmptyList;  ListFull:=false;

  if not UlpAcc(Start,1) then                              { Start iteration }
    begin                                                  {-----------------}
      Y:= Start; WorkList:= EmptyList;
      repeat
        k:= MaxDiamComp(Y);                       { d(Y[k] = max_i d(Y[i])   }
        U[1]:= Y;  U[2]:= Y;                      { Bisect 'Y' with respect  }
        U[1][k].sup:= c[k];  U[2][k].inf:= c[k];  { to component 'k'.        }
        for i:= 1 to 2 do                         {--------------------------}
        begin
          fgEvalH(f, U[i], fU, GradU);           { Compute interval gradient }
                                                 {---------------------------}
          if not MonotonicityTest(GradU) then
          begin                                  { Try centered form to get  }
            fU:= (fC + GradU*(U[i] - c)) ** fU;  { better enclosure of 'f(U)'}
                                                 {---------------------------}
            if (fmax >= inf(fU)) then
            begin
              fghEvalH(f, U[i], fU, GradU, HessU);{ Compute interval Hessian }
                                                  {--------------------------}
              if not ConcavityTest(HessU) then
              begin
                NewtonStep(f,U[i],HessU,V,p);{ Extended interval Newton step }
                                             {-------------------------------}
                for j:=1 to p do
                begin
                  fgEvalH(f, V[j], fV, GradV);   { Compute interval gradient }
                                                 {---------------------------}
                  if not MonotonicityTest(GradV) then
                  begin
                    cV:= mid(V[j]);                       { Try centered form}
                    fEvalH(f,intval(cV),fCV);             { to get better en-}
                    fV:= (fCV + GradV*(V[j] - cV)) ** fV; { closure of 'f(U)'}
                                                          {------------------}
                    if (fmax >= inf(fV)) then                       { Store V}
                    begin                                           {--------}
                      PairY:= MakePair(V[j],inf(fV));
                      Enter(WorkList, VecMat, PairY, ListFull);
                    end;
                  end;
                end; { for j ... }
              end;  { if not ConcavityTest(HessU) }
            end; { if fmax >= ... }
          end; { if not MonotonicityTest(gradU) ... }
        end; { for i ... }
                                                           { Get next 'Y' of }
                                                           { the work list   }
        Bisect:= false;                                    {-----------------}
        while (WorkList <> EmptyList) and (not Bisect) and (not ListFull) do
        begin
          PairY:= Head(WorkList,VecMat);  DelHead(WorkList);

          Y:= Int(PairY);  c:= mid(Y);
          fEvalH(f, intval(c), fC);                           { Compute f(c) }
                                                              {--------------}
          if sup(fC) < fmax then fmax:= sup(fC);
          MultiDelete(WorkList,VecMat,fmax);

          Minimum:= intval(Fyi(PairY),fmax);
                                                { Check termination criteria }
          if (RelDiam(Minimum)<=Epsilon) or (MaxRelDiam(Y)<=Epsilon) then
            Enter(ResultList, VecMat, PairY, ListFull)       { Store 'PairY' }
          else                                               {---------------}
            Bisect:= true;
        end;

      until (not Bisect);

    end { if not UlpAcc(Start,1) }
  else
    begin                                          { Store starting interval }
      fEvalH(f, Start, fY);                        { and interval evaluation }
      Enter(ResultList,VecMat,MakePair(Start,inf(fY)),ListFull);
    end;
                                                { Compute good enclosure of  }
  if ResultList <> EmptyList then               { the global minimum value   }
   Minimum:= intval(Fyi(Head(ResultList,VecMat)),fmax);{---------------------}
end;

{----------------------------------------------------------------------------}
{ Purpose: Reduce the number of boxes in the result list of the algorithm.   }
{ Parameters:                                                                }
{    In/Out : 'ResultList' : stores the candidates for enclosure of a global }
{                            minimizer.                                      }
{             'VecMat'     : matrix for storing the vectors of the list.     }
{ Description:                                                               }
{   The procedure manages the 'ResultList' of pending subintervals that may  }
{   contain global minimizers. Subintervals are removed from the list and    }
{   united with other subintervals if their intersection is not empty.       }
{----------------------------------------------------------------------------}
procedure Compress (var ResultList : PairPtr; var VecMat : imatrix);
var
  PairY1, PairY2 : ivector[0..ub(VecMat,2)];    { Pairs ( Y, inf(f(Y) )      }
  Y1, Y2         : ivector[1..ub(VecMat,2)];    { Interval vectors           }
  Lin, Lhelp     : PairPtr;                     { List of pairs              }
  ListFull       : boolean;
begin
  Lhelp:= ResultList;  ResultList:= EmptyList;
  while (Lhelp <> EmptyList) do
  begin
    Lin:= Lhelp;            { 'Lin'   : list of elements to be checked now   }
    Lhelp:= EmptyList;      { 'Lhelp' : list of elements to be checked again }
    PairY1:= Head(Lin,VecMat);             { Get first element of 'Lin', use }
    Y1:= Int(PairY1);                      { its vector part 'Y1', and re-   }
    DelHead(Lin);                          { move it from the list 'Lin'     }
    while Lin <> EmptyList do
    begin
      PairY2:= Head(Lin,VecMat);           { Get next element of 'Lin', use  }
      Y2:= Int(PairY2);                    { its vector part 'Y2', and check }
                                           { disjointedness of 'Y1' and 'Y2' }
      if Y1 >< Y2 then
        Enter(Lhelp, VecMat, PairY2, ListFull)   { Enter 'PairY2' in 'Lhelp' }
                                                 { for further investigation }
      else
        Y1 := Y1 +* Y2;                          { Set 'Y1' to the union of  }
                                                 { 'Y1' and 'Y2'             }
      DelHead(Lin);                              { Delete 'Y2' in 'Lin'      }
    end;
    PairY1:=MakePair(Y1,Fyi(PairY1));            { Enter (possibly enlarged) }
    Enter(ResultList, VecMat, PairY1, ListFull); { 'Y1' in 'ResultList'      }
  end;
end;

{----------------------------------------------------------------------------}
{ 'MaxNorm' delivers an upper bound of the maximum norm of a symmetric in-   }
{ terval matrix 'H', i.e. the row sum norm or infinity norm.                 }
{----------------------------------------------------------------------------}
function MaxNorm (var H: imatrix) : real;
var
  Nm, MaxNm : real;
  i, j      : integer;
begin
  MaxNm:= 0;
  for i:=lb(H) to ub(H) do
  begin
    Nm:= 0;
    for j:=lb(H) to ub(H) do Nm := Nm +> sup (abs (H[i,j]));
    if Nm > MaxNm then MaxNm:= Nm;
  end;
  MaxNorm:= MaxNm;
end;

{----------------------------------------------------------------------------}
{ 'PosDef' delivers true if it is guaranteed that all real symmetric matri-  }
{ ces enclosed in 'H' are positive definite.                                 }
{----------------------------------------------------------------------------}
function PosDef (var H: imatrix) : boolean;
const
  kmax = 10;  { Maximum number of iterations }
var
  pd         : boolean;
  kappa, eps : real;
  S          : imatrix[lb(H,1)..ub(H,1),lb(H,2)..ub(H,2)];
  Z, U       : ivector[lb(H,1)..ub(H,1)];
  k          : integer;
begin
  kappa:= MaxNorm(H);   S:= Id(H) - H/kappa;
  for k:=lb(Z) to ub(Z) do Z[k]:= intval(-1,1);
  k := 0;  eps:= 0.25;
  repeat
    U:= blow(Z,eps);  Z:= S * U;  pd:= Z in U;
    k:= k+1;  eps:= 8*eps;
  until pd or (k = kmax);
  PosDef:= pd;
end;

{----------------------------------------------------------------------------}
{ Purpose: Execution of a verification step including the use of an epsilon  }
{    inflation.                                                              }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'HessType'.                            }
{    Out    : 'yUnique' : returns 'true' if the verification is successful.  }
{    In/Out : 'y'       : interval enclosure to be verified.                 }
{ Description: This procedure checks the uniqueness of the local minimizer   }
{    enclosed in the interval variable 'y' by a verification step including  }
{    the use of an epsilon inflation of the iterates.                        }
{----------------------------------------------------------------------------}
procedure VerificationStep (function f (x: HTvector) : HessType[0..ub(x,1),
                                                                0..ub(x,2)];
                                 var               y : ivector;
                                 var         yUnique : boolean);
const
  kmax = 10;  { Maximum number of iterations }
var
  fY               : interval;
  yIn, yOld, GradY : ivector[1..ub(y)];
  HessY            : imatrix[1..ub(y),1..ub(y)];
  yp               : imatrix[1..ub(y)+1,1..ub(y)];
  k, p             : integer;
  eps              : real;
begin
  yUnique := (inf(y) = sup(y));               { y is a point interval vector }
  if not yUnique then                         {------------------------------}
  begin
    yIn := y;  yOld := y;  k := 0;  eps:= 0.25;            { Initializations }
                                                           {-----------------}
    while (not yUnique) and (k < kmax) do{ Do kmax loops to achieve inclus.  }
    begin                                {-----------------------------------}
      yOld := blow(y,eps);               { Epsilon inflation of 'y'          }
                                         {-----------------------------------}
      { Perform interval Newton step }
      {------------------------------}
      k := k+1;
      fghEvalH(f, yOld, fY, GradY, HessY);    { Compute gradient and Hessian }
                                              {------------------------------}
      NewtonStep(f, yOld, HessY, yp, p);

      if (p <> 1) then                            { No verification possible }
        k := kmax                                 {--------------------------}
      else if (yp[1] = yOld) then
        eps := eps * 8                            { Increase 'eps'           }
      else                                        {--------------------------}
        begin
          y:= yp[1];  yUnique:= y in yOld;{ Inner inclusion ===> uniqueness  }
        end;                              { of a stationary point            }
    end;                                  {----------------------------------}
    if yUnique then yUnique:= PosDef(HessY);  { Positive definite ===> local }
  end;                                        { minimizer is verified        }
  if not yUnique then y:= yIn;                {------------------------------}
end;

{----------------------------------------------------------------------------}
{ Purpose: Computation of enclosures for all global minimizers and for the   }
{    global minimum value of a twice continuously differentiable multi-di-   }
{    mensional, scalar-valued function.                                      }
{ Parameters:                                                                }
{    In     : 'f'             : objective function, must be declared for the }
{                               'HessType' to enable the internal use of     }
{                               the differentiation arithmetic 'hess_ari'.   }
{             'Start',        : specifies the starting interval vector.      }
{             'Epsilon'.      : specifies the desired relative accuracy      }
{                               (interval diameter) of the result boxes.     }
{    Out    : 'OptiVector'    : stores (row by row) and returns the boxes    }
{                               (enclosures) for the global optimizers of    }
{                               'f'. 'OptiVector' is a vector of interval    }
{                               vectors (ie. an interval matrix).            }
{             'InfoVector'    : stores the corresponding information on the  }
{                               uniqueness of the local optimizers in these  }
{                               enclosures.                                  }
{             'NumberOfOptis' : number of enclosures computed.               }
{             'Minimum'     ' : enclosure for the minimum value.             }
{             'Err'           : error code.                                  }
{ Description:                                                               }
{    The enclosures for the global minimizers of 'f' are computed by calling }
{    procedure 'GlobalOptimize'. Then a verification step is applied.        }
{    The enclosures (boxes) for the global minimizers of 'f' are stored in   }
{    the interval matrix 'OptiVector' row by row, the corresponding          }
{    information on the uniqueness of the local minimizers in these          }
{    enclosures is stored in the Boolean vector 'InfoVector'. The number of  }
{    enclosures computed is returned in the integer variable 'NumberOfOptis'.}
{    The enclosure for the global minimum value is returned in the variable  }
{    'Minimum'. If an error occurs, the value of 'Err' is different from 0.  }
{----------------------------------------------------------------------------}
global procedure AllGOp (function f (x: HTvector) : HessType[0..ub(x,1),
                                                             0..ub(x,2)];
                              var           Start : ivector;
                                          Epsilon : real;
                              var      OptiVector : imatrix;
                              var      InfoVector : bvector;
                              var   NumberOfOptis : integer;
                              var         Minimum : interval;
                              var             Err : integer);
var
  i, k          : integer;
  ResultList, L : PairPtr;
  Start1n       : ivector[1..ub(Start)-lb(Start)+1];
  ResMat        : imatrix[1..ListSize,0..ub(Start)-lb(Start)+1];
  ListFull      : boolean;
  MinEpsilon    : real;
begin
  NumberOfOptis:= 0;
  if (lb(OptiVector) <> 1) then       { Check index bounds of result vectors }
    Err:= lbOptiVecNot1               {--------------------------------------}
  else if (lb(InfoVector) <> 1) then
    Err:= lbInfoVecNot1
  else if (ub(InfoVector) <> ub(OptiVector)) then
    Err:= VecsDiffer
  else
    begin                             { Start global optimization method     }
      Err:= NoError;                  {--------------------------------------}
      MinEpsilon:= succ(1.0) - 1.0;   { Relative machine accuracy (1 ulp)    }
      Start1n:= Start;                { Resize to standard bounds 1..n       }
                                      {--------------------------------------}
      if (Epsilon < MinEpsilon) then Epsilon := MinEpsilon;  { Set 'Epsilon' }
                                                             { to 1 ulp acc. }
      GlobalOptimize(f,Start1n,Epsilon,ResultList,Minimum,ResMat,ListFull);

      NumberOfOptis:= Length(ResultList);

      if NumberOfOptis = 0 then  Err:= NoStatOpti;

      if ListFull then Err:= ListTooSmall;

      if ResultList <> EmptyList then
      begin
        Compress(ResultList,ResMat);            { Reduce the number of boxes }
        NumberOfOptis:= Length(ResultList);     {----------------------------}
      end;
                                             { Check if there are more opti- }
      if ub(OptiVector) < NumberOfOptis then { mizers than storage space     }
      begin                                  {-------------------------------}
        Err:= VecTooSmall;  NumberOfOptis:= ub(OptiVector);
      end;

      L:= ResultList;                            { Verification step for the }
      for i:=1 to NumberOfOptis do               { enclosure intervals       }
      begin                                      {---------------------------}
        OptiVector[i]:= Int(Head(L, ResMat));  L := Next(L);
        VerificationStep(f, OptiVector[i], InfoVector[i]);
      end;

      FreeAll(ResultList);
    end;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
