{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Computing enclosures for all global minimizers and for the global }
{    minimum value of a twice continuously differentiable one-dimensional,   }
{    scalar valued function.                                                 }
{ Method: Bisection method combined with midpoint, monotonicity, concavity   }
{    test and extended interval Newton step.                                 }
{ Global procedures and functions:                                           }
{    procedure AllGOp1(...)       : computes enclosures for all zeros        }
{    function  AllGOp1ErrMsg(...) : delivers an error message text           }
{ Updates:                                                                   }
{    23.11.1995 : Termination criterium changed to ' <= Epsilon'             }
{----------------------------------------------------------------------------}
module gop1;

use
  i_ari,     { Interval arithmetic          }
  xi_ari,    { Extended interval arithmetic }
  lst1_ari,  { List arithmetic              }
  ddf_ari,   { Differentiation arithmetic   }
  i_util,    { Interval utilities           }
  b_util;    { Boolean utilities            }

{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError       = 0;  { No error occurred.                                   }
  lbOptiVecNot1 = 1;  { Lower bound of variable OptiVector is not equal to 1.}
  lbInfoVecNot1 = 2;  { Lower bound of variable InfoVector is not equal to 1.}
  VecsDiffer    = 3;  { Bounds of OptiVector and InfoVector do not match.    }
  VecTooSmall   = 4;  { OptiVector too small. Not all zeros can be stored.   }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function AllGOp1ErrMsg ( Err : integer ) : string;
var
  Msg : string;
begin
  case Err of
    NoError      : Msg := '';
    lbOptiVecNot1: Msg := 'Lower bound of OptiVector is not equal to 1';
    lbInfoVecNot1: Msg := 'Lower bound of InfoVector is not equal to 1';
    VecsDiffer   : Msg := 'Bounds of OptiVector and InfoVector do not match';
    VecTooSmall  : Msg := 'Not all optimizers found. OptiVector is too small';
    else         : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  AllGOp1ErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Purpose: Execution of one extended interval Newton step for the derivative }
{    of function 'f'.                                                        }
{ Parameters:                                                                }
{    In     : 'f'    : must be declared for the 'DerivType' to enable the    }
{                      internal use of the differentiation  arithmetic       }
{                      'ddf_ari'.                                            }
{             'Y'    : specifies the starting interval.                      }
{             'ddfY' : f''(Y), already computed outside of 'NewtonStep'.     }
{    Out    : 'V'    : Pair of intervals V[1] and V[2].                      }
{             'p'    : Number of valid intervals in V (0, 1, or 2).          }
{ Description:                                                               }
{    One extended interval Newton step for 'Y' is executed resulting in 'p'  }
{     interval(s) 'V[1]' and 'V[2]' which can be empty.                      }
{----------------------------------------------------------------------------}
procedure NewtonStep (function f (x: DerivType) : DerivType;
                                       Y, ddfY  : interval;
                           var               V  : ivector;
                           var               p  : integer);
var
  c       : real;
  fC, dfC : interval;
begin
  c:= mid(Y);  dfEval(f, intval(c), fC, dfC);   { Midpoint evaluation of 'f' }
  V:= Y ** (c - dfC div ddfY);                  { Execution of Newton step   }
  if V[1] <> EmptyIntval then p:= 1 else p:= 0; { Fix number of non-empty    }
  if V[2] <> EmptyIntval then p:= p+1;          { intervals                  }
end;                                            {----------------------------}

{----------------------------------------------------------------------------}
{ Purpose: Execution of the global optimization method including a bisection }
{    method, midpoint test, monotonicity test, concavity test, and extended  }
{    interval Newton steps.                                                  }
{ Parameters:                                                                }
{    In     : 'f'          : must be declared for the 'DerivType' to enable  }
{                            the internal use of the differentiation         }
{                            arithmetic 'ddf_ari'.                           }
{             'Start:      : specifies the starting interval.                }
{             'Epsilon'    : specifies the desired relative accuracy         }
{    Out    : 'ResultList' : stores the candidates for enclosure of a global }
{                            minimizer.                                      }
{             'Minimum'    : stores the enclosure of the global minimum      }
{                            value.                                          }
{ Description:                                                               }
{   The procedure manages the list 'L' of pending subintervals that may      }
{   contain global minimizers. Subintervals are removed from the list and    }
{   placed in the accepted list 'ResultList' when they satisfy relative      }
{   error acceptance criteria. Subintervals are also removed from the list   }
{   by the midpoint, monotonicity, concavity tests, or by the interval       }
{   Newton steps. Subintervals are added to the pending list when an element }
{   from the list is bisected or when the extended interval Newton step      }
{   yields two candidate intervals.                                          }
{   'ResultList' returns the list of enclosures of the global minimizers,    }
{   'Minimum' returns the enclosure of the global minimum value.             }
{----------------------------------------------------------------------------}
procedure GlobalOptimize (function f (x: DerivType) : DerivType;
                                              Start : interval;
                                            Epsilon : real;
                               var       ResultList : PairPtr;
                               var          Minimum : interval);

var
  PairY                 : Pair;          { Pair ( Y, inf(f(Y) )              }
  Y, fY, dfY, ddfY, BdP : interval;      { Initial box, f(Y), f'(Y), f''(Y)  }
  U, V                  : ivector[1..2]; { Subboxes of Y and U               }
  fU, fV, fC, fCV, fBdP : interval;      { Function evaluations of f         }
  dfU, dfV, ddfU        : interval;      { Derivative evaluations of f       }
  fmax                  : real;          { Upper bound for minimum value     }
  c, cV                 : real;          { Midpoints of intervals Y and V    }
  WorkList              : PairPtr;       { List of pairs                     }
  i, j, p               : integer;       { Control variables                 }
  Bisect                : boolean;       { Flag for iteration and algorithm  }

begin
  c:= mid(Start);  fEval(f, intval(c), fC);{ Compute upper bound for minimum }
  fmax:= sup(fC);                          {---------------------------------}

  if not UlpAcc(Start,1) then                                 { Start method }
    begin                                                     {--------------}
      Y:= Start;  WorkList:= EmptyList;  ResultList:= EmptyList;

      BdP:= inf(Y);                       { Treat boundary points separately }
      for i:=1 to 2 do                    {----------------------------------}
      begin
        fEval(f, BdP, fBdP);
        if sup(fBdP) < fmax then fmax:= sup(fBdP);
        if fmax >= inf(fBdP) then
          WorkList:= WorkList + MakePair(BdP,inf(fBdP));
        BdP:= sup(Y);
      end;

      repeat                                               { Start iteration }
        U[1]:= intval(inf(Y),c);  U[2]:= intval(c,sup(Y)); { Bisect 'Y'      }
        for i:= 1 to 2 do                                  {-----------------}
        begin
          dfEval(f, U[i], fU, dfU);{ Compute dfu = f'(U)                     }
                                   {-----------------------------------------}
          if (0 in dfU) then       { Monotonicity test: if not 0 in dfU stop }
          begin                    {-----------------------------------------}
            fU:= (fC + dfU * (U[i] - c)) ** fU;       { Centered form 'f(U)' }
                                                      {----------------------}
            if (fmax >= inf(fU)) then
            begin
              ddfEval(f, U[i], fU, dfU, ddfU);  { Compute ddfU = f''(U).     }
                                                {----------------------------}
              if (sup(ddfU) >= 0) then    { Concavity test: if 0 > ddfU stop }
              begin                       {----------------------------------}
                NewtonStep(f,U[i],ddfU,V,p); { Extended interval Newton step }
                                             {-------------------------------}
                for j:=1 to p do
                begin
                  dfEval(f, V[j], fV, dfV);       { Compute first derivative }
                                                  {--------------------------}
                  if (0 in dfV) then { Monotonic. test: if not 0 in dfV stop }
                  begin              {---------------------------------------}
                    cV:= mid(V[j]);                     { Try centered form  }
                    fEval(f, intval(cV), fCV);          { to get better en-  }
                    fV:= (fCV + dfV*(V[j] - cV)) ** fV; { closure of 'f(U)'  }
                                                        {--------------------}
                    if (fmax >= inf(fV)) then                      { Store V }
                      WorkList:= WorkList + MakePair(V[j],inf(fV));{---------}
                  end;
                end; { for j ... }
              end;  { if sup(ddfU) >= 0 }
            end; { if fmax >= ... }
          end; { if 0 in dfU ... }
        end; { for i ... }

        Bisect:= false;                                    { Get next 'Y' of }
        while (WorkList <> EmptyList) and (not Bisect) do  { the work list   }
        begin                                              {-----------------}
          PairY:= Head(WorkList);  DelHead(WorkList);

          Y:= GetInt(PairY);  c:= mid(Y);  fEval(f, intval(c), fC);
          if sup(fC) < fmax then fmax:= sup(fC);
          MultiDelete(WorkList,fmax);

          Minimum:= intval(GetFyi(PairY),fmax);
          if (RelDiam(Minimum) <= Epsilon) or (RelDiam(Y) <= Epsilon) then
            ResultList:= ResultList + PairY           { Checking termination }
          else                                        { criterion            }
            Bisect:= true;                            {----------------------}
        end; { while }

      until (not Bisect);

    end { if not UlpAcc(Start,1) }
  else
    begin                                          { Store starting interval }
      fEval(f, Start, fY);                         {-------------------------}
      ResultList:= EmptyList + MakePair(Start,inf(fY));
    end;
                                                    { Compute good enclosure }
  Minimum:= intval(GetFyi(Head(ResultList)),fmax);  { of the global minimum  }
end;                                                {------------------------}

{----------------------------------------------------------------------------}
{ Purpose: Execution of a verification step including the use of an epsilon  }
{    inflation.                                                              }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'DerivType'.                           }
{    Out    : 'yUnique' : returns 'true' if the verification is successful.  }
{    In/Out : 'y'       : interval enclosure to be verified.                 }
{ Description: This procedure checks the uniqueness of the local minimizer   }
{    enclosed in the interval variable 'y' by a verification step including  }
{    the use of an epsilon inflation of the iterates.                        }
{----------------------------------------------------------------------------}
procedure VerificationStep (function f (x: DerivType) : DerivType;
                                 var                y : interval;
                                 var          yUnique : boolean);
const
  kmax = 10;  { Maximum number of iterations }

var
  yIn, yOld, fm, dfm, fy, dfy, ddfy : interval;
  m, eps                            : real;
  k                                 : integer;
begin
  yUnique := (y.inf = y.sup);                        { y is a point interval }
  if not yUnique then                                {-----------------------}
  begin
    yIn := y;  k := 0;  eps:= 0.25;     { Initializations                    }
  end;                                  {------------------------------------}

  while (not yUnique) and (k < kmax) do { Do kmax loops to achieve inclusion }
  begin                                 {------------------------------------}
    yOld:= blow(y, eps);                { Epsilon inflation of 'y'           }
    ddfEval(f, yOld, fy, dfy, ddfy);    {------------------------------------}

    if 0 >= inf(ddfy) then           { No verification of a minimum possible }
      k := kmax                      {---------------------------------------}
    else
      begin                                   { Perform interval Newton step }
        k:= k+1;  m:= mid(yOld);              {------------------------------}
        dfEval(f, intval(m), fm, dfm);
        y := (m - dfm / ddfy);

        if y >< yOld then                         { No verification possible }
          k:= kmax                                {--------------------------}
        else
          begin
            yUnique:= y in yOld;    { Inner inclusion ===> unique zero of f' }
            y      := y ** yOld;    { Intersection with old value            }
            if y = yOld then        {----------------------------------------}
              eps := eps * 8        { Increase the value of 'eps'            }
          end;                      {----------------------------------------}
      end;
  end;
  if not yUnique then y := yIn;
end;

{----------------------------------------------------------------------------}
{ Purpose: Computation of enclosures for all global minimizers and for the   }
{    global minimum value of a twice continuously differentiable one-dimen-  }
{    sional, scalar-valued function.                                         }
{ Parameters:                                                                }
{    In     : 'f'             : objective function, must be declared for the }
{                               'DerivType' to enable the internal use of    }
{                               the differentiation arithmetic 'ddf_ari'.    }
{             'Start',        : specifies the starting interval.             }
{             'Epsilon'.      : specifies the desired relative accuracy      }
{                               (interval diameter) of the result intervals. }
{    Out    : 'OptiVector'    : stores and returns the enclosures for the    }
{                               global optimizers of 'f'.                    }
{             'InfoVector'    : stores the corresponding information on the  }
{                               uniqueness of the local optimizers in these  }
{                               enclosures.                                  }
{             'NumberOfOptis' : returns the number of enclosures computed.   }
{             'Minimum'     ' : returns the enclosure for the minimum value. }
{             'Err'           : returns an error code.                       }
{ Description:                                                               }
{    The enclosures for the global minimizers of 'f' are computed by calling }
{    procedure 'GlobalOptimize'. Then a verification step is applied.        }
{    The enclosures for the global minimizers of 'f' are stored in the       }
{    interval vector 'OptiVector', the corresponding information on the      }
{    uniqueness of the local minimizers in these enclosures is stored in the }
{    Boolean vector 'InfoVector'. The number of enclosures computed is       }
{    returned in the integer variable 'NumberOfOptis'. The enclosure for the }
{    global minimum value is returned in the variable 'Minimum'.             }
{    If an error occurs, the value of 'Err' is different from 0.             }
{----------------------------------------------------------------------------}
global procedure AllGOp1 (function f (x: DerivType) : DerivType;
                                              Start : interval;
                                            Epsilon : real;
                               var       OptiVector : ivector;
                               var       InfoVector : bvector;
                               var    NumberOfOptis : integer;
                               var          Minimum : interval;
                               var              Err : integer);
var
  i             : integer;
  MinEpsilon    : real;
  ResultList, L : PairPtr;
begin
  NumberOfOptis:= 0;
  if (lb(OptiVector) <> 1) then       { Check index bounds of result vectors }
    Err:= lbOptiVecNot1               {--------------------------------------}
  else if (lb(InfoVector) <> 1) then
    Err:= lbInfoVecNot1
  else if (ub(InfoVector) <> ub(OptiVector)) then
    Err:= VecsDiffer
  else
    begin                                { Start global optimization method  }
      Err:= NoError;                     {-----------------------------------}
      MinEpsilon:= succ(1.0) - 1.0;      { Relative machine accuracy (1 ulp) }
                                         {-----------------------------------}
      if (Epsilon < MinEpsilon) then Epsilon := MinEpsilon;  { Set 'Epsilon' }
                                                             { to 1 ulp acc. }
      GlobalOptimize(f,Start,Epsilon,ResultList,Minimum);    {---------------}

      NumberOfOptis:= Length(ResultList);
                                             { Check if there are more opti- }
      if ub(OptiVector) < NumberOfOptis then { mizers than storage space     }
      begin                                  {-------------------------------}
        Err:= VecTooSmall;  NumberOfOptis:= ub(OptiVector);
      end;

      L:= ResultList;
                                                 { Verification step for the }
      for i:=1 to NumberOfOptis do               { enclosure intervals       }
      begin                                      {---------------------------}
        OptiVector[i]:= GetInt(Head(L));  L:= Next(L);
        VerificationStep(f, OptiVector[i], InfoVector[i]);
      end;

      FreeAll(ResultList);
    end;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
