{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Utilities of type 'interval'.                                     }
{ Global functions:                                                          }
{    function AbsMin(...)  : Smallest absolute value of an interval.         }
{    function AbsMax(...)  : Greatest absolute value of an interval.         }
{    function RelDiam(...) : Relative diameter of an interval.               }
{    function UlpAcc(...)  : To check whether the width of an interval is    }
{                            less than a certain number of ulps (ulp = unit  }
{                            in the last place of the mantissa).             }
{    function Power(...)   : Exponentiation by an integer for intervals.     }
{    function Pi           : Returns an enclosure of pi.                     }
{ Updates:                                                                   }
{    14.11.1995 : Modification of 'RelDiam' wrt. denormalized numbers and    }
{                 overflow problems                                          }
{----------------------------------------------------------------------------}
module i_util;     { Interval Utilities }
                   { -        ----      }
use
  i_ari;           { Interval arithmetic }

var
  LocalPi     : interval;   { Local variable to store an enclosure of pi }
  MinNormReal : real;       { Smallest postive normalized machine number }

global function AbsMin ( x : interval ) : real;        { Absolute minimum of }
begin                                                  { an interval         }
  if (0 in x) then                                     {---------------------}
    AbsMin := 0
  else if (inf(x) > 0) then
    AbsMin := inf(x)
  else
    AbsMin := -sup(x);
end;

global function AbsMax ( x : interval ) : real;        { Absolute maximum of }
var                                                    { an interval         }
  a, b : real;                                         {---------------------}
begin
  a := abs(inf(x));  b := abs(sup(x));
  if (a > b) then  AbsMax := a  else  AbsMax := b;
end;

global function RelDiam ( x : interval ) : real;       { Relative diameter   }
var AM, D : real;                                      { of an interval      }
begin                                                  {---------------------}
  D := diam(x);
  if D = 0 then
    RelDiam := 0
  else
    begin
      AM := AbsMin(x);
      if (AM <= MinNormReal) or (AM/D <= MinNormReal) then
        RelDiam := D
      else
        RelDiam := D /> AM;
    end;
end;

{----------------------------------------------------------------------------}
{ Checks whether the width of the interval 'x' is less or equal to 'n' ulp.  }
{ An ulp is an abbreviation for: units in the last place of the mantissa.    }
{----------------------------------------------------------------------------}
global function UlpAcc ( x : interval; n : integer ) : boolean;
var
  i       : integer;
  Infimum : real;
begin
  Infimum := inf(x);
  for i := 1 to n do Infimum := succ(Infimum);
  UlpAcc := (Infimum >= sup(x));
end;

{----------------------------------------------------------------------------}
{ Purpose: The local function 'Power()' is used to compute a lower or an     }
{    upper bound for the power function with real argument and integer       }
{    exponent, respectively.                                                 }
{ Parameters:                                                                }
{    In  : 'x'       : real argument.                                        }
{          'n'       : integer exponent.                                     }
{          'RndMode' : rounding mode,                                        }
{                      (-1 = downwardly directed, +1 = upwardly directed)    }
{ Description:                                                               }
{    This function is used to speed up the interval power function defined   }
{    below. The exponentiation is reduced to multiplications using the       }
{    binary shift method. Depending on 'n', this function is up to 40 times  }
{    as fast as the standard power function for real argument and real       }
{    exponent. However, its accuracy is less than one ulp (unit in the last  }
{    place of the mantissa) since about log2(n) multiplications are executed }
{    during computation. Since directed roundings are antisymmetric, one     }
{    gets                                                                    }
{                                                                            }
{       down(x^n) = -up((-x)^n)   and   up(x^n) = -down((-x)^n)              }
{                                                                            }
{    for x < 0 and odd n, where 'down' and 'up' denote the downwardly and    }
{    upwardly directed roundings, respectively.                              }
{----------------------------------------------------------------------------}
function Power ( x : real; n, RndMode : integer ) : real;
var                          { Signals change of the rounding mode }
  ChangeRndMode : boolean;   { for x < 0 and odd n                 }
  p, z          : real;
begin
  ChangeRndMode := ((x < 0) and odd(n));
  if ChangeRndMode then
    begin z := -x; RndMode := -RndMode; end
  else
    z := x;

  p := 1;                                    { Note: Seperate while-loops    }
  case RndMode of                            { used to gain speed at runtime }
    -1 : while (n > 0) do                    {-------------------------------}
         begin
           if odd(n) then p := p *< z;
           n := n div 2;
           if (n > 0) then z := z *< z;
         end;
    +1 : while (n > 0) do
         begin
           if odd(n) then p := p *> z;
           n := n div 2;
           if (n > 0) then z := z *> z;
         end;
  end;
  if ChangeRndMode then  Power := -p  else  Power := p;
end;

{----------------------------------------------------------------------------}
{ Purpose: This version of the function 'Power()' is used to compute an      }
{    enclosure for the power function with interval argument and integer     }
{    exponent.                                                               }
{ Parameters:                                                                }
{    In  : 'x' : interval argument.                                          }
{          'n' : integer exponent.                                           }
{ Description:                                                               }
{    In general, this implementation does not deliver a result of maximum    }
{    accuracy, but it is about 30-40 times faster than the standard power    }
{    function for interval arguments and interval exponents. The resulting   }
{    interval has a width of approximately 2*log2(n) ulps. Since x^n is      }
{    considered as a monomial, we define x^0 := 1. For negative exponents    }
{    and 0 in 'x', the division at the end of the function will cause a      }
{    runtime error (division by zero).                                       }
{----------------------------------------------------------------------------}
global function Power ( x : interval; n : integer ) : interval;
var
  m            : integer;
  Lower, Upper : real;
begin
  if (n = 0) then
    Power := 1
  else
    begin
      if (n > 0) then  m := n  else  m := -n;

      if (0 < x.inf) or odd(m) then
        begin  Lower := Power(x.inf,m,-1);  Upper := Power(x.sup,m,+1);  end
      else if (0 > x.sup) then
        begin  Lower := Power(x.sup,m,-1);  Upper := Power(x.inf,m,+1);  end
      else
        begin  Lower := 0;  Upper := Power(AbsMax(x),m,+1);  end;

      if (n > 0) then
        Power := intval(Lower,Upper)
      else                                     { Propagates a 'division by }
        Power := 1 / intval(Lower,Upper);      { zero' error if 0 in 'x'.  }
    end;
end;

{----------------------------------------------------------------------------}
{ The function 'Pi' returns an enclosure of pi. To save computation time, pi }
{ is computed only once in the module initialization part by evaluating      }
{ 4*arctan(intval(1)). Thus, this function only returns the value of the     }
{ local variable 'LocalPi'.                                                  }
{----------------------------------------------------------------------------}
global function Pi : interval;
begin
  Pi := LocalPi;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  LocalPi := 4*arctan(intval(1));    { Computes an enclosure of pi }
  MinNormReal := comp(0.5, -1021);   { Computes smallest normalized number }
end.
