{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of a list arithmetic used in connection with an        }
{    interval bisection method in global optimization for storing pairs of   }
{    an interval vector and a real value.                                    }
{ Method: Representing pairs of an interval vector and a real value by a     }
{    interval vector with additional 0-component storing the real value.     }
{    Overloading of functions and procedures for the data type 'PairPtr'.    }
{ Global types, functions, and procedures:                                   }
{    types      PairPtr, PairElmt: list of pairs                             }
{    functions  MakePair         : transfer function for pairs               }
{               Int, Fyi         : access functions for pairs                }
{               Next, Head       : access functions for lists                }
{               Length           : access function to length of list         }
{               EmptyList        : delivers an empty list                    }
{    procedures FreeAll          : free complete list                        }
{               MultiDelete      : deletes several elements in a list        }
{               DelHead          : deletes first element of a list           }
{----------------------------------------------------------------------------}
module lst_ari;

use
  i_ari,     { Interval arithmetic               }
  mvi_ari;   { Interval matrix/vector arithmetic }

{----------------------------------------------------------------------------}
{ Global type definitions                                                    }
{----------------------------------------------------------------------------}
global type                  { List of pairs of an interval vector 'y' and   }
  PairPtr  = ^PairElmt;      { the real value 'inf(f(y))'.  The pair is re-  }
  PairElmt = record          { presented by an interval vector with index    }
               P : integer;  { bounds 0..n and the 0-component contains the  }
               N : PairPtr;  { real value. Within the list, each pair-vector }
             end;            { is represented by the index 'P' and indirect- }
                             { ly stored in 'VecMat[P]', where 'VecMat' is a }
                             { storage matrix provided by the routine using  }
                             { this module. 'N' means NEXT, the pointer to   }
                             { the next element of the list.                 }
                             {-----------------------------------------------}
{----------------------------------------------------------------------------}
{ Local variable storing list of free elements (automatic garbage recycling) }
{----------------------------------------------------------------------------}
var
  FreeList : PairPtr;
  IndexNo  : integer;

{----------------------------------------------------------------------------}
{ Procedures for generating and freeing of list elements (pairs)             }
{----------------------------------------------------------------------------}
procedure NewPP (var pp: PairPtr);    { 'NewPP' generates a new list element }
begin                                 { or gets one from 'FreeList'.         }
  if FreeList = nil then              {--------------------------------------}
    begin
      new(pp);  pp^.N:= nil;  IndexNo:= IndexNo+1;  pp^.P:= IndexNo;
    end
  else
    begin
      pp:= FreeList;  FreeList:= FreeList^.N;  pp^.N:= nil;
    end;
end;

procedure Free (var pp: PairPtr);           { 'Free' enters one element of a }
begin                                       { list in the 'FreeList'.        }
  if pp <> nil then                         {--------------------------------}
  begin
      pp^.N:= FreeList;  FreeList := pp;  pp:= nil;
  end;
end;

global procedure FreeAll (var List: PairPtr);{ 'FreeAll' enters all elements }
var  H : PairPtr;                            { of 'List' in the 'FreeList'.  }
begin                                        {-------------------------------}
  if List <> nil then
  begin
    H:= List;
    while H^.N <> nil do H:= H^.N;
    H^.N:= FreeList;  FreeList:= List;  List:= nil;
  end;
end;

{----------------------------------------------------------------------------}
{ Transfer and access functions for pairs                                    }
{----------------------------------------------------------------------------}
global function MakePair (var int: ivector; fyi: real) : ivector[0..ub(int)];
var i : integer;                                             { Generate pair }
begin                                                        {---------------}
  MakePair[0]:= fyi;
  for i:=1 to ub(int) do MakePair[i]:= int[i];
end;

global function Int (var Pair: ivector) : ivector[1..ub(Pair)];
var i : integer;                     { Get interval vector component of pair }
begin                                {---------------------------------------}
  for i:=1 to ub(Pair) do Int[i]:= Pair[i];
end;

global function Fyi (var Pair: ivector) : real; { Get real component of pair }
begin                                           {----------------------------}
  Fyi:= Pair[0].inf;
end;

{----------------------------------------------------------------------------}
{ Functions, and procedures for lists of pairs                               }
{----------------------------------------------------------------------------}
{ Global function 'EmptyList' representing an empty list of pairs.           }
{----------------------------------------------------------------------------}
global function EmptyList : PairPtr;
begin
  EmptyList := nil;
end;

{----------------------------------------------------------------------------}
{ Function 'Enter' enters 'Pair' the list 'List' (by storing the vector in   }
{ the matrix 'VecMat') in such a way that after entering, one of the four    }
{ following condition holds:                                                 }
{ 1) Fyi(O) <= Fyi(Pair) < Fyi(Q),                                           }
{ 2)           Fyi(Pair) < Fyi(Q) and 'Pair' is the first element of 'List', }
{ 3) Fyi(O) <= Fyi(Pair)          and 'Pair' is the last  element of 'List', }
{ 4)                                  'Pair' is the only  element of 'List', }
{ where 'O' is the preceding and 'Q' is the succeeding element of 'Pair' in  }
{ the resulting list.  If the list (ie. the storage matrix 'VecMat') is full,}
{ the parameter 'Full' returns true.                                         }
{----------------------------------------------------------------------------}
global procedure Enter (var List: PairPtr; var VecMat: imatrix;
                        var Pair: ivector; var Full  : boolean);
var
  H, HN            : PairPtr;
  ready, alreadyIn : boolean;
begin
  Full:= false;
  if (List = nil) then                          { List is empty, so new list }
    begin                                       { only contains 'Pair'.      }
      NewPP(H);                                 {----------------------------}
      if H^.P > ub(VecMat) then
        Full:= true
      else
        VecMat[H^.P]:= Pair;
      H^.N:= nil;  List:= H;
    end
  else if (Fyi(VecMat[List^.P]) > Fyi(Pair)) then { 'Pair' becomes new first }
    begin                                         { element of the list.     }
      NewPP(H);                                   {--------------------------}
      if H^.P > ub(VecMat) then
        Full:= true
      else
        VecMat[H^.P]:= Pair;
      H^.N:= List;  List:= H;
    end
  else
    begin
      H:= List;  HN:= H^.N;  ready:= false;
      alreadyIn:= (Int(VecMat[H^.P]) = Int(Pair));

      while not (ready or alreadyIn) do         { Search for the right       }
      begin                                     { position to enter 'Pair'.  }
        if (HN = nil) then                      {----------------------------}
          ready:= true
        else if (Fyi(VecMat[HN^.P]) > Fyi(Pair)) then
          ready:= true
        else
          begin
            H:= HN;  HN:= H^.N;  alreadyIn:= (Int(VecMat[H^.P]) = Int(Pair));
          end;
      end;

      if not alreadyIn then
      begin
        NewPP(H^.N);  H:= H^.N;                     { Enter 'Pair' between H }
        if H^.P > ub(VecMat) then                   { and HN. Return List.   }
          Full:= true                               {------------------------}
        else
          VecMat[H^.P]:= Pair;
        H^.N:= HN;
      end;
    end;
end;

{----------------------------------------------------------------------------}
{ 'MultiDelete' deletes all elements 'P' in 'List' for which the condition   }
{ 'Fyi(P) > fmax' holds.  This procedure assumes that the 'fyi' components   }
{ of the list elements are sorted in increasing order (see function 'Enter').}
{----------------------------------------------------------------------------}
global procedure MultiDelete (var List: PairPtr; var MV: imatrix; fmax: real);
var
  DelPrev, Del : PairPtr;
  ready        : boolean;
begin
  if (List <> nil) then
  begin
    if (Fyi(MV[List^.P]) > fmax) then            { All list elements fulfill }
      begin                                      { 'Fyi(P) > fmax'.          }
        Del:= List;  List:= nil;                 {---------------------------}
      end
    else
      begin
        DelPrev:= List;  Del:= DelPrev^.N;  ready:= (Del=nil);

        while not ready do
        begin
          if (Del = nil) then
            ready := true
          else if (Fyi(MV[Del^.P]) > fmax) then
            begin
              ready:= true;  DelPrev^.N:= nil;
            end
          else
            begin
              DelPrev:= Del;  Del:= Del^.N;
            end;
        end;
      end;
    FreeAll(Del);
  end;
end;

global function Next (List: PairPtr) : PairPtr;   { Sets list pointer to the }
begin                                             { next list element        }
  Next:= List^.N;                                 {--------------------------}
end;

global function Head (List: PairPtr; var MV: imatrix): ivector[0..ub(MV,2)];
                                                { Delivers first pair of the }
begin                                           { list, i.e. the pair P with }
  Head:= MV[List^.P];                           { the smallest value P[0].   }
end;                                            {----------------------------}

global procedure DelHead (var List: PairPtr);   { Deletes the first pair of  }
var                                             { the List.                  }
  Del : PairPtr;                                {----------------------------}
begin
  Del := List;  List:= List^.N;  Free(Del);
end;

global function Length (List: PairPtr) : integer;  { 'Length' delivers the   }
var  i : integer;                                  { number of elements in   }
begin                                              { list 'List'.            }
  i:= 0;                                           {-------------------------}
  while List <> nil do
  begin
    i:= succ(i); List:= List^.N;
  end;
  Length := i;
end;

{----------------------------------------------------------------------------}
{ Module initialization                                                      }
{----------------------------------------------------------------------------}
begin
  FreeList := nil;         { List of freed elements which can be used again  }
  IndexNo  := 0;           { Index of last row of storage matrix used        }
end.                       {-------------------------------------------------}
